/*****************************************************************************
*
*      xr16l78x.c :	XR16L788/784 Family Multiport Serial Driver.
*
*      Copyright (C) 2001,2002  Exar Corp.	(www.exar.com)
*	   Copyright (C) 1991, 1992, Linus Torvalds
*      Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997,
*					 1998, 1999, Theodore Ts'o
*
*      This code is loosely based on the Linux serial driver, written by
*      Linus Torvalds, Theodore T'so and others.
*
*      This program is free software; you can redistribute it and/or modify
*      it under the terms of the GNU General Public License as published by
*      the Free Software Foundation; either version 2 of the License, or
*      (at your option) any later version.
*
*      This program is distributed in the hope that it will be useful,
*      but WITHOUT ANY WARRANTY; without even the implied warranty of
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*      GNU General Public License for more details.
*
*      You should have received a copy of the GNU General Public License
*      along with this program; if not, write to the Free Software
*      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*	   XR16l78x Family Multiport Serial Driver
*	   for			: LINUX		(Tested with RedHat 7.0 and 7.1)
*	   date			: February, 2004
*	   version		: 2.0	
*	   08/12/2001		: XR16L788, XR16L784 ISA boards support
*
* This version includes RS485 support via an ioctl (EXAR_SET_RS485)
*	   
********************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include "exser.h"

//The EXAR serial driver boot-time initialization code!
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0) )
static int exar_init_module(void)
#else
int init_module (void)
#endif
{
	int	ret;

	if (verbose)
		printk("Loading module exser ...\n");
	    printk("Default IRQ = 5, IO = 0x500\n");
    ret = exser_init();
	if (verbose)
		printk("Done.\n");
	return (ret);
}

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0) )
static void exar_cleanup_module(void)
#else
void cleanup_module(void)
#endif
{
	int i, err = 0;

	if (verbose)
		printk("Unloading module exser ...\n");
	if ((err |= tty_unregister_driver(&exvar_cdriver)))
        printk("Couldn't unregister EXAR callout driver\n");
    if ((err |= tty_unregister_driver(&exvar_sdriver)))
        printk("Couldn't unregister EXAR serial driver\n");

    for(i=0; i<EXSER_BOARDS; i++)
	{
		if(exsercfg[i].board_type == -1)
			continue;
        else
		{
            free_irq(exsercfg[i].irq, &exvar_table[i*EXSER_PORTS_PER_BOARD]);
			exser_clearled(&exsercfg[i]);
        }
    }    
	if (verbose)
		printk("Done.\n");
}

static int exser_init(void)
{
	int		i, m, retval, b;
	int		n, index;
    int     ret1, ret2;
	unsigned char		busnum, devnum;
    struct exser_hwconf     hwconf;
	struct pci_dev	* pcidev = NULL;

    printk("EXAR serial multiport driver version %s\n", EXSER_VERSION);

	hwconf.brdid = 1;
	//Initialize the tty_driver structure
    memset(&exvar_sdriver, 0, sizeof(struct tty_driver));
    exvar_sdriver.magic = TTY_DRIVER_MAGIC;
    exvar_sdriver.name = "ttyM";
    exvar_sdriver.major = ttymajor;
    exvar_sdriver.minor_start = 0;
    exvar_sdriver.num = EXSER_PORTS + 1;
    exvar_sdriver.type = TTY_DRIVER_TYPE_SERIAL;
    exvar_sdriver.subtype = SERIAL_TYPE_NORMAL;
    exvar_sdriver.init_termios = tty_std_termios;
    exvar_sdriver.init_termios.c_cflag = B9600|CS8|CREAD|HUPCL|CLOCAL;
    exvar_sdriver.flags = TTY_DRIVER_REAL_RAW;
    exvar_sdriver.refcount = &exvar_refcount;
    exvar_sdriver.table = exvar_tty;
    exvar_sdriver.termios = exvar_termios;
    exvar_sdriver.termios_locked = exvar_termios_locked;

    exvar_sdriver.open = exser_open;
    exvar_sdriver.close = exser_close;
    exvar_sdriver.write = exser_write;
    exvar_sdriver.put_char = exser_put_char;
    exvar_sdriver.flush_chars = exser_flush_chars;
    exvar_sdriver.write_room = exser_write_room;
    exvar_sdriver.chars_in_buffer = exser_chars_in_buffer;
    exvar_sdriver.flush_buffer = exser_flush_buffer;
    exvar_sdriver.ioctl = exser_ioctl;
    exvar_sdriver.throttle = exser_throttle;
    exvar_sdriver.unthrottle = exser_unthrottle;
    exvar_sdriver.set_termios = exser_set_termios;
    exvar_sdriver.stop = exser_stop;
    exvar_sdriver.start = exser_start;
    exvar_sdriver.hangup = exser_hangup;

	//The callout device is just like normal device except for
	//major number and the subtype code.
    exvar_cdriver = exvar_sdriver;
    exvar_cdriver.name = "cum";
    exvar_cdriver.major = calloutmajor;
    exvar_cdriver.subtype = SERIAL_TYPE_CALLOUT;

	printk("Tty devices major number = %d, callout devices major number = %d\n", ttymajor, calloutmajor);

    exvar_diagflag = 0;
    memset(exvar_table, 0, EXSER_PORTS * sizeof(struct exser_struct));
    memset(&exvar_log, 0, sizeof(struct exser_log));

#if !defined(XR158_PCI) && !defined(XR154_PCI)
	ioaddr[0] = IO;
#else
	ioaddr[0] = 0x0;
#endif
	m = 0;
	// Start to finding ISA boards from module arg
    for (b=0; b<EXSER_BOARDS && m<EXSER_BOARDS; b++) 
	{
		int cap;
	    if ( !(cap=ioaddr[b]) )
	        continue;
		retval = get_isr_conf(cap, &hwconf);

	    if (retval != 0) 
		{
			printk("Found EXAR %s board (I/O=0x%x)\n",
                    exser_brdname[hwconf.board_type],
				    ioaddr[b]);
		}		
	    if (retval <= 0) 
		{
			if (retval == EXSER_ERR_IRQ)
				printk("Invalid interrupt number, board not configured\n");
            else if (retval == EXSER_ERR_IRQ_CONFLIT)
				printk("Invalid interrupt number, board not configured\n");
            else if (retval == EXSER_ERR_VECTOR)
				printk("Invalid interrupt vector, board not configured\n");
            else if (retval == EXSER_ERR_IOADDR)
				printk("Invalid I/O address, board not configured\n");
			continue;
	    }   
	    hwconf.pciInfo.busNum = 0;
	    hwconf.pciInfo.devNum = 0;
        if(exser_initbrd(m, &hwconf)<0)
			continue;
	    exser_getcfg(m, &hwconf);		  
	    m++;
	}

//start to finding pciboard here
#ifdef CONFIG_PCI
#if (LINUX_VERSION_CODE < VERSION_CODE(2, 1, 0))
	if (pcibios_present()) {
#else
	if (pci_present()) {
#endif
		n = sizeof (exser_pcibrds)/sizeof (exser_pciinfo);
		index = 0;
		b = 0;
		while (b < n) 
		{
			if (pcibios_find_device(exser_pcibrds[b].vendor_id,
                        exser_pcibrds[b].device_id,
						index,
						&busnum,
						&devnum) != 0) 
			{
				b++;
				index = 0;
				continue;
			}
		
			hwconf.pciInfo.busNum = busnum;
			hwconf.pciInfo.devNum = devnum;
            printk("Found EXAR %s board(BusNo=%d, DevNo=%d)\n", exser_brdname[exser_pcibrds[b].board_type-1], 
								busnum,devnum >> 3);
			index++;
            if (m >= EXSER_BOARDS) 
			{
				printk("Too many EXAR multiport serial boards find (maximum %d), board not configured\n", EXSER_BOARDS);
			}
			else 
			{
                pcidev = pci_find_device( exser_pcibrds[b].vendor_id, exser_pcibrds[b].device_id, pcidev);
				hwconf.brdid = index;
                retval = exser_get_PCI_conf(busnum, devnum,
                                        exser_pcibrds[b].board_type, &hwconf, pcidev);
				if (retval < 0) 
				{
					if (retval == EXSER_ERR_IRQ)
						printk("Invalid interrupt number, board not configured\n");
                    else if (retval == EXSER_ERR_IRQ_CONFLIT)
						printk("Invalid interrupt number, board not configured\n");
                    else if (retval == EXSER_ERR_VECTOR)
						printk("Invalid interrupt vector, board not configured\n");
                    else if (retval == EXSER_ERR_IOADDR)
						printk("Invalid I/O address, board not configured\n");
					continue;
				}                              
			if(exser_initbrd(m, &hwconf)<0)
				continue;
            exser_getcfg(m, &hwconf);
			m++;					
			}
		}
	}
#endif
	
	for(i=m; i<EXSER_BOARDS; i++)
	{
		exsercfg[i].board_type = -1;
	}

    ret1 = 0;
    ret2 = 0;
    if ( !(ret1=tty_register_driver(&exvar_sdriver)) )
	{
		if ( !(ret2=tty_register_driver(&exvar_cdriver)) )
		{
			return 0;
        }
		else
		{
            tty_unregister_driver(&exvar_sdriver);
            printk("Couldn't install EXAR Smartio family callout driver !\n");
        }
    }
	else
        printk("Couldn't install EXAR Smartio family driver !\n");

    if(ret1 || ret2)
	{
	    for(i=0; i<EXSER_BOARDS; i++)
		{
		    if(exsercfg[i].board_type == -1)
	            continue;
            else
			{
                free_irq(exsercfg[i].irq, &exvar_table[i*EXSER_PORTS_PER_BOARD]);
            }
        }
        return -1;
    }    
	return(0);
}

static int exser_initbrd(int board, struct exser_hwconf * hwconf)
{
    struct exser_struct *   info;
	unsigned long	flags;
    int     retval;
	int		i, n;
	
    n = board*EXSER_PORTS_PER_BOARD;
    info = &exvar_table[n];
	for (i=0; i<hwconf->ports; i++, n++, info++) 
	{
		if (verbose) 
		{
			printk("  ttyM%d/cum%d at 0x%04x ", n, n, hwconf->ioaddr[i]);
			if (hwconf->baud_base[i] == 115200)
		    	printk(" max. baud rate up to 115200 bps.\n");
			else
		    	printk(" max. baud rate up to 921600 bps.\n");
		}
		INIT_LIST_HEAD(&info->tqueue.list);
		info->port = n;
		info->base = hwconf->ioaddr[i];
		info->irq = hwconf->irq;
		info->vector = hwconf->vector;
		info->vectormask = hwconf->vector_mask;
		info->rx_trigger = 14;
		info->baud_base = hwconf->baud_base[i];
		info->flags = ASYNC_SHARE_IRQ;
		info->type = hwconf->uart_type;
		if ( (info->type == PORT_16450) || (info->type == PORT_8250) )
		    info->xmit_fifo_size = 1;
		else
		    info->xmit_fifo_size = 16;
		info->custom_divisor = hwconf->baud_base[i]*16;
		info->close_delay = 5*HZ/10;
		info->closing_wait = 30*HZ;
        info->tqueue.routine = exser_do_softint;
		info->tqueue.data = info;
        info->callout_termios = exvar_cdriver.init_termios;
        info->normal_termios = exvar_sdriver.init_termios;
		init_waitqueue_head(&info->open_wait);
		init_waitqueue_head(&info->close_wait);
		init_waitqueue_head(&info->delta_msr_wait);
	}
//Allocate the IRQ if necessary
    n = board*EXSER_PORTS_PER_BOARD;
    info = &exvar_table[n];
    save_flags(flags); 
	cli();
#if defined(XR158_PCI) || defined(XR154_PCI)
	retval = request_irq(hwconf->irq, exser_interrupt, SA_SHIRQ,
                                 "exser", info);
#else
	retval = request_irq(hwconf->irq, exser_interrupt, SA_INTERRUPT, //IRQ_T(info),
                             "exser", info);
#endif
	if (retval) 
	{
	    restore_flags(flags);
        printk("Board %d: %s", board, exser_brdname[hwconf->board_type-1]);
	    printk("Request irq fail, IRQ (%d) may be conflit with another device.\n", info->irq);
	    return(retval);
	}
	restore_flags(flags);

    return 0;
}

static void exser_getcfg(int board, struct exser_hwconf * hwconf)
{
    exsercfg[board] = *hwconf;
}

static int get_isr_conf(int cap, struct exser_hwconf * hwconf)
{
	int		i;
	int		iTemp = 0;
	unsigned char	scratch, scratch2;

	iTemp = byte_read(cap + ISA_XR_DVID);
	if(iTemp == ISA_XR784)
		hwconf->board_type = ISA_BOARD_XR784;
	else
		hwconf->board_type = ISA_BOARD_XR788;
	hwconf->ports = exser_numports[hwconf->board_type - 1];
	hwconf->irq = IRQ & 0xff;
	for (i=0; i<hwconf->ports; i++)
	{
		hwconf->ioaddr[i] = cap + 0x10*i;
	}
	hwconf->vector = cap + 0x80;
	hwconf->uart_type = PORT_16550A;
	hwconf->vector_mask = 0;
    for (i=0; i<hwconf->ports; i++) 
	{
		hwconf->vector_mask |= (1<<i);
        hwconf->baud_base[i] = 921600;	
		scratch2 = byte_read(hwconf->ioaddr[i] + UART_LCR) & (~UART_LCR_DLAB);
		byte_write(scratch2 | UART_LCR_DLAB, hwconf->ioaddr[i] + UART_LCR);
		byte_write(0, hwconf->ioaddr[i] + UART_EFR);
		byte_write(scratch2, hwconf->ioaddr[i] + UART_LCR);
		byte_write(UART_FCR_ENABLE_FIFO, hwconf->ioaddr[i] + UART_FCR);
		scratch = byte_read(hwconf->ioaddr[i] + UART_IIR);
	}

	return(1);
}

static int exser_get_PCI_conf(int busnum, int devnum, int board_type, 
				  struct exser_hwconf * hwconf, struct pci_dev * pcidev)
{
	int		i;
	unsigned int	val, ioaddress;
	u32		pci_io_base;

	hwconf->board_type = board_type;
        hwconf->ports = exser_numports[board_type-1];

	pci_io_base = pci_resource_start(pcidev, 0);	
	ioaddress = (unsigned int)ioremap((pci_io_base & PCI_BASE_ADDRESS_MEM_MASK), 0x1000);

	for (i=0; i<hwconf->ports; i++) 
	{
		hwconf->ioaddr[i] = ioaddress + 0x200*i;
	}
	hwconf->vector = ioaddress + 0x80;

	pcibios_read_config_dword(busnum, devnum, PCI_INTERRUPT_LINE, &val);
	if (val == 0xffffffff)
        return (EXSER_ERR_IRQ);
	else 
		hwconf->irq = val & 0xff;
	
	hwconf->uart_type = PORT_16550A;
	hwconf->vector_mask = 0;
    for (i=0; i<hwconf->ports; i++) 
	{
		hwconf->vector_mask |= (1<<i);
        hwconf->baud_base[i] = 921600;
	}

	exser_setled(hwconf);
	return(0);
}

static void exser_do_softint(void * private_)
{
    struct exser_struct *   info = (struct exser_struct *)private_;
	struct tty_struct *	tty;

	tty = info->tty;
	if ( !tty )
	    return;
#ifdef  __SMP__
	lock_kernel();
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2, 1, 0))
    if ( clear_bit(EXSER_EVENT_TXLOW, &info->event) ) 
	{
	    if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		 tty->ldisc.write_wakeup )
			(tty->ldisc.write_wakeup)(tty);
	    wake_up_interruptible(&tty->write_wait);
	}
    if ( clear_bit(EXSER_EVENT_HANGUP, &info->event) ) 
	{
		tty_hangup(tty);
	}
#else
    if ( test_and_clear_bit(EXSER_EVENT_TXLOW, &info->event) ) 
	{
	    if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		      tty->ldisc.write_wakeup )
		      (tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
    if ( test_and_clear_bit(EXSER_EVENT_HANGUP, &info->event) ) 
	{
		tty_hangup(tty);
	}
#endif
#ifdef  __SMP__
	unlock_kernel();
#endif
}

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
static int exser_open(struct tty_struct * tty, struct file * filp)
{
    struct exser_struct *   info;
	int		retval, line;
	unsigned long		page;

	line = PORTNO(tty);
    if (line == EXSER_PORTS)
	    return(0);
    if ( (line < 0) || (line > EXSER_PORTS) )
	    return(-ENODEV);
    info = exvar_table + line;
	if ( !info->base )
	    return(-ENODEV);

	info->count++;
	tty->driver_data = info;
	info->tty = tty;

    if ( !exvar_tmp_buf ) 
	{
	    page = get_free_page(GFP_KERNEL);
	    if ( !page )
			return(-ENOMEM);
        if ( exvar_tmp_buf )
			free_page(page);
	    else
            exvar_tmp_buf = (unsigned char *) page;
	}

	//Start up serial port
    retval = exser_startup(info);
	if (retval)
	    return(retval);

    retval = exser_block_til_ready(tty, filp, info);
	if (retval)
	    return(retval);

	if ( (info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS) ) 
	{
	    if ( tty->driver.subtype == SERIAL_TYPE_NORMAL )
			*tty->termios = info->normal_termios;
	    else
			*tty->termios = info->callout_termios;
        exser_change_speed(info, 0);
	}
	info->session = current->session;
	info->pgrp = current->pgrp;

	MOD_INC_USE_COUNT;
	printk("OPEN...\n");
	return(0);
}

/*
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 */
static void exser_close(struct tty_struct * tty, struct file * filp)
{
    struct exser_struct * info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;
	unsigned long	timeout;

	printk("exser: close\n");	
    if ( PORTNO(tty) == EXSER_PORTS )
	    return;
	if ( !info )
	    return;

	save_flags(flags);
	cli();

	if ( tty_hung_up_p(filp) ) 
	{
	    MOD_DEC_USE_COUNT;
		restore_flags(flags);
	    return;
	}
	if ( (tty->count == 1) && (info->count != 1) ) 
	{
	    /*
	     * Uh, oh.	tty->count is 1, which means that the tty
	     * structure will be freed.  Info->count should always
	     * be one in these conditions.  If it's greater than
	     * one, we've got real problems, since it means the
	     * serial port won't be shutdown.
	     */
        printk("exser_close: bad serial port count, tty->count is 1, "
		   "info->count is %d\n", info->count);
	    info->count = 1;
	}
	if ( --info->count < 0 ) 
	{
        printk("exser_close: bad serial port count for ttys%d: %d\n",
		         info->port, info->count);
	    info->count = 0;
	}
	if (info->count) 
	{
	    MOD_DEC_USE_COUNT;
		restore_flags(flags);
	    return;
	}
	info->flags |= ASYNC_CLOSING;
	restore_flags(flags);

	//Save the termios structure, since this port may have
	//separate termios for callout and dialin.
	if ( info->flags & ASYNC_NORMAL_ACTIVE )
		info->normal_termios = *tty->termios;
	if ( info->flags & ASYNC_CALLOUT_ACTIVE )
	    info->callout_termios = *tty->termios;

	//Now we wait for the transmit buffer to clear; and we notify
	//the line discipline to only process XON/XOFF characters.
	tty->closing = 1;
	if ( info->closing_wait != ASYNC_CLOSING_WAIT_NONE )
	    tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	info->IER &= ~UART_IER_RLSI;
	if ( info->flags & ASYNC_INITIALIZED ) 
	{
	    byte_write(info->IER, info->base + UART_IER);
	    /*
	     * Before we drop DTR, make sure the UART transmitter
	     * has completely drained; this is especially
	     * important if there is a transmit FIFO!
	     */
	    timeout = jiffies + HZ;
	    while ( !(byte_read(info->base + UART_LSR) & UART_LSR_TEMT) ) 
		{
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(5);
			if ( jiffies > timeout )
				break;
	    }
	}
    exser_shutdown(info);
	if (tty->driver.flush_buffer)
	    tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
	    tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) 
	{
	    if (info->close_delay) 
		{
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
	    }
	    wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE |
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
}

static int exser_write(struct tty_struct * tty, int from_user,
		       const unsigned char * buf, int count)
{
	int	c, total = 0;
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;
	
//	printk("exser: writing\n");

    if ( !tty || !info->xmit_buf || !exvar_tmp_buf )
	    return(0);

	if (from_user)
		down(&exvar_tmp_buf_sem);
	save_flags(flags);
	while ( 1 ) 
	{
	    cli();
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			       SERIAL_XMIT_SIZE - info->xmit_head));
	    if (c <= 0)
		break;

	    if (from_user) 
		{
			copy_from_user(exvar_tmp_buf, buf, c);
			c = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			       SERIAL_XMIT_SIZE - info->xmit_head));
            memcpy(info->xmit_buf + info->xmit_head, exvar_tmp_buf, c);
	    } 
		else
		{
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
		}
	    info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE - 1);
	    info->xmit_cnt += c;
	    restore_flags(flags);
	    buf += c;
	    count -= c;
	    total += c;
	}
	if (from_user)
		up(&exvar_tmp_buf_sem);
	if ( info->xmit_cnt && !tty->stopped && !tty->hw_stopped &&
	     !(info->IER & UART_IER_THRI) ) 
	{
	    info->IER |= UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);
	}
	restore_flags(flags);
	return(total);
}

static void exser_put_char(struct tty_struct * tty, unsigned char ch)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;

	if ( !tty || !info->xmit_buf )
	    return;

	save_flags(flags);
	cli();
	if (info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) 
	{
	    restore_flags(flags);
	    return;
	}

	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE - 1;
	info->xmit_cnt++;

	restore_flags(flags);
}

static void exser_flush_chars(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;

	if ( info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
	     !info->xmit_buf )
	    return;

	save_flags(flags);
	cli();
	info->IER |= UART_IER_THRI;
	byte_write(info->IER, info->base + UART_IER);
	restore_flags(flags);
}

static int exser_write_room(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	int	ret;

	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
	    ret = 0;
	return(ret);
}

static int exser_chars_in_buffer(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	return(info->xmit_cnt);
}

static void exser_flush_buffer(struct tty_struct * tty)
{
	unsigned long flags;
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
//	printk("exser: flushing buffers\n");
	save_flags(flags); 
	cli();
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	restore_flags(flags);
	wake_up_interruptible(&tty->write_wait);
	if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	     tty->ldisc.write_wakeup )
	    (tty->ldisc.write_wakeup)(tty);
}

static int exser_ioctl(struct tty_struct * tty, struct file * file,
		       unsigned int cmd, unsigned long arg)
{
	int	error;
    struct exser_struct *   info = (struct exser_struct *)tty->driver_data;
	int	retval;
	struct async_icount	cprev, cnow;
	struct serial_icounter_struct *p_cuser;     
	unsigned long	templ, flags;
	
//	printk("exser: doing ioctl: %d\n", cmd);
    if ( PORTNO(tty) == EXSER_PORTS )
		return(exser_ioctl_special(cmd, arg));
	if ( (cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) &&
	     (cmd != TIOCGICOUNT) ) 
	{
	    if (tty->flags & (1 << TTY_IO_ERROR)) 
		{
			printk("HERE1..\n");
			return(-EIO);
	    }
	}
	switch (cmd) {
	case TCSBRK:	
		//SVID version: non-zero arg --> no break
	    retval = tty_check_change(tty);
	    if (retval)
			return(retval);
	    tty_wait_until_sent(tty, 0);
	    if ( !arg )
            exser_send_break(info, HZ/4);           //  1/4 second
	    return(0);
	case TCSBRKP:	
		//support for POSIX tcsendbreak()
	    retval = tty_check_change(tty);
	    if (retval)
			return(retval);
	    tty_wait_until_sent(tty, 0);
        exser_send_break(info, arg ? arg*(HZ/10) : HZ/4);
	    return(0);
	case TIOCGSOFTCAR:
	    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(long));
	    if (error)
			return(error);
	    put_to_user(C_CLOCAL(tty) ? 1 : 0, (unsigned long *)arg);
	    return(0);
	case TIOCSSOFTCAR:
	    error = verify_area(VERIFY_READ, (void *)arg, sizeof(long));
	    if (error)
			return(error);
	    get_from_user(templ,(unsigned long *)arg);
	    arg = templ;
	    tty->termios->c_cflag = ((tty->termios->c_cflag & ~CLOCAL) |
				    (arg ? CLOCAL : 0));
	    return(0);
	case TIOCMGET:
	    error = verify_area(VERIFY_WRITE, (void *)arg,
				sizeof(unsigned int));
	    if (error)
			return(error);
        return(exser_get_modem_info(info, (unsigned int *)arg));
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
            return(exser_set_modem_info(info, cmd, (unsigned int *)arg));
	case TIOCGSERIAL:
	    error = verify_area(VERIFY_WRITE, (void *)arg,
				sizeof(struct serial_struct));
	    if (error)
			return(error);
        return(exser_get_serial_info(info, (struct serial_struct *)arg));
	case TIOCSSERIAL:
	    error = verify_area(VERIFY_READ, (void *)arg,
				sizeof(struct serial_struct));
	    if (error)
			return(error);
        return(exser_set_serial_info(info, (struct serial_struct *)arg));
	case TIOCSERGETLSR: 
		//Get line status register
	    error = verify_area(VERIFY_WRITE, (void *)arg,
				sizeof(unsigned int));
	    if (error)
			return(error);
	    else
            return(exser_get_lsr_info(info, (unsigned int *)arg));
	/*
	 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change

	 * - mask passed in arg for lines of interest
	 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
	 * Caller should use TIOCGICOUNT to see which one it was
	 */
	case TIOCMIWAIT:
	    save_flags(flags);
	    cli();
		//note the counters on entry
	    cprev = info->icount;   
	    restore_flags(flags);
	    while ( 1 ) 
		{
		interruptible_sleep_on(&info->delta_msr_wait);
		//see if a signal did it
		if (signal_pending(current))
		    return(-ERESTARTSYS);
		save_flags(flags); 
		cli();
		//atomic copy
		cnow = info->icount;	
		restore_flags(flags);
		if ( cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
		     cnow.dcd == cprev.dcd && cnow.cts == cprev.cts ) 
		{
		    printk("HERE2...\n");
		    return(-EIO);		
		}
		if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
		     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
		     ((arg & TIOCM_CD)	&& (cnow.dcd != cprev.dcd)) ||
		     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) 
		{
		    return(0);
		}
		cprev = cnow;
	    }
	    //NOTREACHED
	/*
	 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
	 * Return: write counters to the user passed counter struct
	 * NB: both 1->0 and 0->1 transitions are counted except for
	 *     RI where only 0->1 is counted.
	 */
	case TIOCGICOUNT:
	    error = verify_area(VERIFY_WRITE, (void *)arg,
				sizeof(struct serial_icounter_struct));
	    if (error)
			return(error);
	    save_flags(flags); 
		cli();
	    cnow = info->icount;
	    restore_flags(flags);
	    p_cuser = (struct serial_icounter_struct *) arg;
	    put_to_user(cnow.cts, &p_cuser->cts);
	    put_to_user(cnow.dsr, &p_cuser->dsr);
	    put_to_user(cnow.rng, &p_cuser->rng);
	    put_to_user(cnow.dcd, &p_cuser->dcd);
	    return(0);
    case EXAR_HighSpeedOn:
	    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if (error)
			return(error);
	    put_to_user(info->baud_base != 115200 ? 1 : 0, (int *)arg);
	    return(0);

	case EXAR_READ_REG:
	    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(struct exser_read_reg));
	    if (error)
			return(error);
        else 
		{
			return(exser_read_reg(info, (struct exser_read_reg *)arg));
	    }
	    break;

    case EXAR_WRITE_REG:
        error = verify_area(VERIFY_READ, (void *)arg, sizeof(struct exser_read_reg));
        if (error)
			return(error);
        else 
		{
            return(exser_write_reg(info, (struct exser_read_reg *)arg));
        }
        break;

    case EXAR_SET_RS485:
      exser_set_autoRS485(info, 1); // 2nd param will be nonzero if need to set
      break;
        

	default:
	//	printk("exser: IOCTL default cmd=%x\n", cmd);
	    return(-ENOIOCTLCMD);
	}
	return(0);
}

static int exser_read_reg(struct exser_struct * info, struct exser_read_reg * arg)
{
    unsigned char   result;
	unsigned int	reg;
	int		error;
    unsigned long flags;

   	error = verify_area(VERIFY_READ, (void *)arg->reg, sizeof(struct exser_read_reg));
    if (error)
		return(error);
    get_from_user(reg, &(arg->reg));
      
	save_flags(flags); 
	cli();
    result = byte_read(info->base + reg);
    restore_flags(flags);
//	printk("exser: Read base=0x%08x, Reg=0x%08x, value=0x%08x\n", info->base, reg, result); 
    put_to_user(result, &(arg->regvalue));
    return(0);
}

static int exser_write_reg(struct exser_struct * info, struct exser_read_reg * arg)
{
    unsigned int reg, value;

    get_from_user(reg, &(arg->reg));
    get_from_user(value, &(arg->regvalue));
    byte_write(value, info->base + reg); 
//  printk("exser: write base=0x%08x, Reg=0x%08x, value=0x%08x\n", info->base, reg, value);
    return(0);
}

// Feature control Register
#define UART_FCTR 0x08

static int exser_set_autoRS485(struct exser_struct * info, unsigned int flag )
{
    unsigned char oldFCTR, oldMSR, oldMCR;

    // RTS is chosen by default. To choose DTR make oldMCR |= 0x04;
    oldMCR = byte_read(info->base + UART_MCR);
    oldMCR &= 0xfb;
    byte_write(oldMCR, info->base + UART_MCR);

    oldFCTR = byte_read(info->base + UART_FCTR);
    if(flag) // set autoRS485
    {
      oldFCTR |= 0x20;
      byte_write(oldFCTR, info->base + UART_FCTR);
      oldMSR = 0x0; // add your delay from 0x10 to 0xF0 (bit times)
      byte_write(oldMSR, info->base + UART_MSR);
    }
    else
    {
      oldFCTR &= 0xDF;       // !0x20
      byte_write(oldFCTR, info->base + UART_FCTR);
      oldMSR = 0x0;
      byte_write(oldMSR, info->base + UART_MSR);
    }
    return(0);
}

static int exser_ioctl_special(unsigned int cmd, unsigned long arg)
{
	int	error, i, result, status;

//	printk("exser: IOCTL_SPECIAL start\n");

	switch (cmd) {
        case EXAR_GET_CONF:
	    error = verify_area(VERIFY_WRITE, (void *)arg, 
                            sizeof(struct exser_hwconf)*4);
	    if (error)
			return(error);
        copy_to_user((struct exser_hwconf *)arg, exsercfg, 
                            sizeof(struct exser_hwconf)*4);
	    return 0;


        case EXAR_GET_MAJOR:
		error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if (error)
			return(error);
	    copy_to_user((int*)arg, &ttymajor, sizeof(int));
        return 0;

        case EXAR_GET_CUMAJOR:
	    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if (error)
			return(error);
	    copy_to_user((int*)arg, &calloutmajor, sizeof(int));
            return 0;

        case EXAR_CHKPORTENABLE:
	    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(long));
	    if (error)
			return(error);
	    result = 0;
        for (i=0; i<EXSER_PORTS; i++) 
		{
			if ( exvar_table[i].base )
		    result |= (1 << i);
	    }
	    put_to_user(result, (unsigned long *)arg);
	    return(0);

        case EXAR_GETDATACOUNT:
	    error = verify_area(VERIFY_WRITE, (void *)arg,
                          sizeof(struct exser_log));
	    if (error)
			return(error);
        copy_to_user((struct exser_log *)arg, &exvar_log, sizeof(exvar_log));
	    return(0);

        case EXAR_GETMSTATUS:
	    error = verify_area(VERIFY_WRITE, (void *)arg,
                          sizeof(struct exser_mstatus) * EXSER_PORTS);
	    if (error)
			return(error);
        for(i=0; i<EXSER_PORTS; i++)
		{
			GMStatus[i].ri = 0;
            if ( !exvar_table[i].base )
			{
				GMStatus[i].dcd = 0;
                GMStatus[i].dsr = 0;
                GMStatus[i].cts = 0;
				continue;
            }

            if ( !exvar_table[i].tty || !exvar_table[i].tty->termios )
				GMStatus[i].cflag=exvar_table[i].normal_termios.c_cflag;
            else
                GMStatus[i].cflag = exvar_table[i].tty->termios->c_cflag;
                
            status = byte_read(exvar_table[i].base + UART_MSR);
            if(status  & 0x80)	//UART_MSR_DCD
				GMStatus[i].dcd = 1;
            else
                GMStatus[i].dcd = 0;

            if(status  & 0x20)	//UART_MSR_DSR
                GMStatus[i].dsr = 1;
            else
                GMStatus[i].dsr = 0;    

            if(status  & 0x10)	//UART_MSR_CTS
                GMStatus[i].cts = 1;
            else
                GMStatus[i].cts = 0;
		}
            copy_to_user((struct exser_mstatus *)arg, GMStatus,
                          sizeof(struct exser_mstatus) * EXSER_PORTS);
            return 0;
	default:
	return(-ENOIOCTLCMD);
	}
	return(0);
}

//This routine is called by the upper-layer tty layer to signal that
//incoming characters should be throttled.
static void exser_throttle(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;
//	printk("exser: throttle\n");
	if ( I_IXOFF(tty) ) 
	{
	    info->x_char = STOP_CHAR(tty);
	    save_flags(flags);
	    cli();
	    byte_write(info->IER, 0);
	    info->IER |= UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);	//force Tx interrupt
	    restore_flags(flags);
	}

	if ( info->tty->termios->c_cflag & CRTSCTS ) 
	{
	    info->MCR &= ~UART_MCR_RTS;
	    save_flags(flags);
	    cli();
	    byte_write(info->MCR, info->base + UART_MCR);
	    restore_flags(flags);
	}
}

static void exser_unthrottle(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;

	if ( I_IXOFF(tty) ) 
	{
	    if (info->x_char)
		{
			info->x_char = 0;
		}
	    else 
		{
			info->x_char = START_CHAR(tty);
			save_flags(flags);
			cli();
			byte_write(info->IER, 0);
			//force Tx interrupt
			info->IER |= UART_IER_THRI;		
			byte_write(info->IER, info->base + UART_IER);
			restore_flags(flags);
	    }
	}

	if ( info->tty->termios->c_cflag & CRTSCTS ) 
	{
	    info->MCR |= UART_MCR_RTS;
	    save_flags(flags);
	    cli();
	    byte_write(info->MCR, info->base + UART_MCR);
	    restore_flags(flags);
	}
}

static void exser_set_termios(struct tty_struct * tty,
			      struct termios * old_termios)
{
	struct exser_struct *info = (struct exser_struct *)tty->driver_data;
//	printk("exser: set_termios\n");

	if ( (tty->termios->c_cflag != old_termios->c_cflag) ||
	     (RELEVANT_IFLAG(tty->termios->c_iflag) !=
	      RELEVANT_IFLAG(old_termios->c_iflag)) ) 
	{
		exser_change_speed(info, old_termios);
		if ( (old_termios->c_cflag & CRTSCTS) &&
	     	     !(tty->termios->c_cflag & CRTSCTS) ) 
		{
			tty->hw_stopped = 0;
            exser_start(tty);
		}
	}
//Handle sw stopped
	if ( (old_termios->c_iflag & IXON) &&
     	     !(tty->termios->c_iflag & IXON) ) 
	{
		tty->stopped = 0;
        exser_start(tty);
	}
}

/*
 * exser_stop() and exser_start()
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 */
static void exser_stop(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;
	
//	printk("exser: stop\n");
	save_flags(flags);
	cli();
	if ( info->IER & UART_IER_THRI ) 
	{
	    info->IER &= ~UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);
	}
	restore_flags(flags);
}

static void exser_start(struct tty_struct * tty)
{
    struct exser_struct *info = (struct exser_struct *)tty->driver_data;
	unsigned long	flags;
	
//	printk("exser: start");
	save_flags(flags);
	cli();
	if ( info->xmit_cnt && info->xmit_buf &&
	     !(info->IER & UART_IER_THRI) ) 
	{
	    info->IER |= UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);
	}
	restore_flags(flags);
}

//This routine is called by tty_hangup() when a hangup is signaled.
void exser_hangup(struct tty_struct * tty)
{
    struct exser_struct * info = (struct exser_struct *)tty->driver_data;
//	printk("exser: hangup\n");
    exser_flush_buffer(tty);
    exser_shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

//This is the serial driver's generic interrupt routine
static void exser_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
	int			status, i;
    struct exser_struct *   info;
    struct exser_struct *   port;
	int			max, irqbits, bits, msr;
	int			pass_counter = 0;

//  printk("exser: interrupt\n"); 
	port = 0;
    for(i=0; i<EXSER_BOARDS; i++)
	{
		if(dev_id == &(exvar_table[i*EXSER_PORTS_PER_BOARD]))
		{
			port = dev_id;
            break;
        }
    }
    if(i==EXSER_BOARDS)
		return;      
	if(port==0)
        return;      
    max = exser_numports[exsercfg[i].board_type-1];
	while ( 1 ) 
	{
		irqbits = byte_read(port->vector) & port->vectormask;
/*   if ( irqbits == port->vectormask )
		break;
*/
		for ( i=0, bits=1; i<max; i++, irqbits |= bits, bits <<= 1 ) 
		{
/*		if ( irqbits == port->vectormask )
		    break;
*/
			if ( !(bits & irqbits) )
				continue;
			info = port + i;
			if ( !info->tty ||
		          (byte_read(info->base + UART_IIR) & UART_IIR_NO_INT) )
				continue;
			status = byte_read(info->base + UART_LSR) & info->read_status_mask;
			if (status & UART_LSR_DR)
				exser_receive_chars(info, &status);
			msr = byte_read(info->base + UART_MSR);
			if (msr & UART_MSR_ANY_DELTA)
				exser_check_modem_status(info, msr);
			if (status & UART_LSR_THRE) 
			{
/*
		    if ( info->x_char || (info->xmit_cnt > 0) )
*/
				exser_transmit_chars(info);
			}
		}
		if ( pass_counter++ > EXSER_ISR_PASS_LIMIT ) 
		{
#if 0
			printk("EXAR multiport serial driver interrupt loop break\n");
#endif
			break;
		}
	}
}

static _INLINE_ void exser_receive_chars(struct exser_struct * info,
					 int * status)
{
	struct tty_struct *	tty = info->tty;
	unsigned char	ch;
	int		ignored = 0;
	int		cnt = 0;
	
//	printk("exser: receiving chars\n");
	do {
	    ch = byte_read(info->base + UART_RX);
	    if ( *status & info->ignore_status_mask ) 
		{
			if (++ignored > 100)
				break;
	    } 
		else 
		{
			if (tty->flip.count >= TTY_FLIPBUF_SIZE)
				break;
			tty->flip.count++;
			if (*status & UART_LSR_SPECIAL) 
			{
				if (*status & UART_LSR_BI) 
				{
					*tty->flip.flag_buf_ptr++ = TTY_BREAK;
					if (info->flags & ASYNC_SAK)
					do_SAK(tty);
				} 
				else if (*status & UART_LSR_PE) 
				{
					*tty->flip.flag_buf_ptr++ = TTY_PARITY;
				} 
				else if (*status & UART_LSR_FE) 
				{
					*tty->flip.flag_buf_ptr++ = TTY_FRAME;
				} 
				else if (*status & UART_LSR_OE) 
				{
					*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
				} 
				else
					*tty->flip.flag_buf_ptr++ = 0;
			} else
				*tty->flip.flag_buf_ptr++ = 0;
			*tty->flip.char_buf_ptr++ = ch;
			cnt++;
		}
	    *status = byte_read(info->base + UART_LSR) & info->read_status_mask;
	} while ( *status & UART_LSR_DR );
    exvar_log.rxcnt[info->port] += cnt;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	queue_task(&tty->flip.tqueue, &tq_timer);
	mark_bh(SERIAL_BH);
#else
	queue_task_irq_off(&tty->flip.tqueue, &tq_timer);
#endif
}

static _INLINE_ void exser_transmit_chars(struct exser_struct * info)
{
	int	count, cnt;

//	printk("exser: transmitting chars\n");
	if (info->x_char) 
	{
	    byte_write(info->x_char, info->base + UART_TX);
	    info->x_char = 0;
        exvar_log.txcnt[info->port]++;
	    return;
	}
	if ( (info->xmit_cnt <= 0) || info->tty->stopped ||
	     info->tty->hw_stopped ) 
	{
	    info->IER &= ~UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);
	    return;
	}

	cnt = info->xmit_cnt;
	count = info->xmit_fifo_size;
	do {
	    byte_write(info->xmit_buf[info->xmit_tail++], info->base + UART_TX);
	    info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE - 1);
	    if ( --info->xmit_cnt <= 0 )
			break;
	} while ( --count > 0 );
    exvar_log.txcnt[info->port] += (cnt - info->xmit_cnt);
	if (info->xmit_cnt < WAKEUP_CHARS) 
	{
        set_bit(EXSER_EVENT_TXLOW, &info->event);
		//tq_serial is the queue to use
		queue_task(&info->tqueue, &tq_serial);
		mark_bh(SERIAL_BH);
	}

	if (info->xmit_cnt <= 0) 
	{
	    info->IER &= ~UART_IER_THRI;
	    byte_write(info->IER, info->base + UART_IER);
	}
}

static _INLINE_ void exser_check_modem_status(struct exser_struct * info,
					      int status)
{
//	printk("exser: check_modem_status");
	//update input line counters
	if (status & UART_MSR_TERI)
	    info->icount.rng++;
	if (status & UART_MSR_DDSR)
	    info->icount.dsr++;
	if (status & UART_MSR_DDCD)
	    info->icount.dcd++;
	if (status & UART_MSR_DCTS)
	    info->icount.cts++;
	wake_up_interruptible(&info->delta_msr_wait);

	if ( (info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD) ) 
	{
	    if (status & UART_MSR_DCD)
		{
			wake_up_interruptible(&info->open_wait);
		}
	    else if ( !((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		      (info->flags & ASYNC_CALLOUT_NOHUP)) )
		{
            set_bit(EXSER_EVENT_HANGUP, &info->event);
		}  
	       //tq_serial is our BH queue
	 queue_task(&info->tqueue,&tq_serial);
	 mark_bh(SERIAL_BH);
	}
	if (info->flags & ASYNC_CTS_FLOW) 
	{
	    if (info->tty->hw_stopped) 
		{
			if (status & UART_MSR_CTS) 
			{
				info->tty->hw_stopped = 0;
				info->IER |= UART_IER_THRI;
				byte_write(info->IER, info->base + UART_IER);
                set_bit(EXSER_EVENT_TXLOW, &info->event);
	      	    queue_task(&info->tqueue, &tq_serial);
				mark_bh(SERIAL_BH);
			}
	    } 
		else 
		{
			if ( !(status & UART_MSR_CTS) ) 
			{
				info->tty->hw_stopped = 1;
				info->IER &= ~UART_IER_THRI;
				byte_write(info->IER, info->base + UART_IER);
			}
	    }

	}
}

static int exser_block_til_ready(struct tty_struct * tty, struct file * filp,
                                 struct exser_struct * info)
{
	DECLARE_WAITQUEUE(wait, current);
	int	retval;
	int	do_clocal = 0;
	unsigned long	flags;
	
	//If the device is in the middle of being closed, then block
	//until it's done, and then try again.
	if ( tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING) ) 
	{
	    if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
	    if (info->flags & ASYNC_HUP_NOTIFY)
			return(-EAGAIN);
	    else
			return(-ERESTARTSYS);
#else
	    return(-EAGAIN);
#endif
	}
	
	//If this is a callout device, then just make sure the normal
	//device isn't being used.
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) 
	{
	    if (info->flags & ASYNC_NORMAL_ACTIVE)
			return(-EBUSY);
	    if ( (info->flags & ASYNC_CALLOUT_ACTIVE) &&
				(info->flags & ASYNC_SESSION_LOCKOUT) &&
				(info->session != current->session) )
			return(-EBUSY);
	    if ( (info->flags & ASYNC_CALLOUT_ACTIVE) &&
				(info->flags & ASYNC_PGRP_LOCKOUT) &&
				(info->pgrp != current->pgrp) )
			return(-EBUSY);
	    info->flags |= ASYNC_CALLOUT_ACTIVE;
	    return(0);
	}

	//If non-blocking mode is set, or the port is not enabled,
	//then make the check up front and then exit.
	if ( (filp->f_flags & O_NONBLOCK) ||
	     (tty->flags & (1 << TTY_IO_ERROR)) ) 
	{
	    if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return(-EBUSY);
	    info->flags |= ASYNC_NORMAL_ACTIVE;
	    return(0);
	}
	
	if (info->flags & ASYNC_CALLOUT_ACTIVE) 
	{
	    if (info->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} 
	else 
	{
	    if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
     * exser_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
	save_flags(flags); 
	cli();
	if ( !tty_hung_up_p(filp) )
	    info->count--;
	restore_flags(flags);
	info->blocked_open++;
	while ( 1 ) 
	{
	    save_flags(flags);
	    cli();
	    if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) )
			byte_write(byte_read(info->base + UART_MCR) | UART_MCR_DTR | UART_MCR_RTS,
						info->base + UART_MCR);
	    restore_flags(flags);
	    current->state = TASK_INTERRUPTIBLE;
	    if ( tty_hung_up_p(filp) || !(info->flags & ASYNC_INITIALIZED) ) 
		{
#ifdef SERIAL_DO_RESTART
			if ( info->flags & ASYNC_HUP_NOTIFY )
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
#else
			retval = -EAGAIN;
#endif
			break;
	    }
	    if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) &&
					!(info->flags & ASYNC_CLOSING) &&
					(do_clocal || (byte_read(info->base + UART_MSR) & UART_MSR_DCD)) )
			break;
	    if ( signal_pending(current) ) 
		{
			retval = -ERESTARTSYS;
			break;
	    }
	    schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if ( !tty_hung_up_p(filp) )
	    info->count++;
	info->blocked_open--;
	if (retval)
	    return(retval);
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return(0);
}

static int exser_startup(struct exser_struct * info)
{
	unsigned long	flags;
	unsigned long	page;

//	printk("exser: startup\n");
	page = get_free_page(GFP_KERNEL);
	if ( !page )
	    return(-ENOMEM);

	save_flags(flags);
	cli();

	if (info->flags & ASYNC_INITIALIZED) 
	{
	    free_page(page);
	    restore_flags(flags);
	    return(0);
	}

	if ( !info->base || !info->type ) 
	{
	    if ( info->tty )
			set_bit(TTY_IO_ERROR, &info->tty->flags);
	    free_page(page);
	    restore_flags(flags);
	    return(0);
	}
	if (info->xmit_buf)
	    free_page(page);
	else
	    info->xmit_buf = (unsigned char *)page;

	//Clear the FIFO buffers and disable them
    //(they will be reenabled in exser_change_speed())
	if (info->xmit_fifo_size == 16)
	    byte_write((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT),
		 info->base + UART_FCR);

	//At this point there's no way the LSR could still be 0xFF;
	//if it is, then bail out, because there's likely no UART here.
	if (byte_read(info->base + UART_LSR) == 0xff) 
	{
	    restore_flags(flags);
	    if (suser()) 
		{
			if (info->tty)
				set_bit(TTY_IO_ERROR, &info->tty->flags);
			return(0);
		} 
		else 
		{
			return(-ENODEV);
		}
	}
         
	//Clear the interrupt registers.
	byte_read(info->base + UART_LSR);
	byte_read(info->base + UART_RX);
	byte_read(info->base + UART_IIR);
	byte_read(info->base + UART_MSR);
	//Now, initialize the UART
	byte_write(UART_LCR_WLEN8, info->base + UART_LCR);	/* reset DLAB */
	info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	byte_write(info->MCR, info->base + UART_MCR);
	//Finally, enable interrupts
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;
	byte_write(info->IER, info->base + UART_IER); /* enable interrupts */
	//And clear the interrupt registers again for luck.
	byte_read(info->base + UART_LSR);
	byte_read(info->base + UART_RX);
	byte_read(info->base + UART_IIR);
	byte_read(info->base + UART_MSR);

#if (LINUX_VERSION_CODE < VERSION_CODE(2, 1, 0))
	if (info->tty)
	    clear_bit(TTY_IO_ERROR, &info->tty->flags);
#else
	if (info->tty)
	    test_and_clear_bit(TTY_IO_ERROR, &info->tty->flags);
#endif
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	//and set the speed of the serial port
    exser_change_speed(info, 0);
//rs485 : uncomment the below to enable RS485 by default
    //exser_set_autoRS485(info, 1); //2nd param will be non-zero for TRUE
//end rs485    
	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
	return(0);
}

//This routine will shutdown a serial port; interrupts maybe disabled, and
//DTR is dropped if the hangup on close termio flag is on.
static void exser_shutdown(struct exser_struct * info)
{
	unsigned long	flags;

//	printk("exser: shutdown\n");
	if ( !(info->flags & ASYNC_INITIALIZED) )
	    return;

	save_flags(flags);
	//Disable interrupts
	cli();			
	//clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	//here so the queue might never be waken up
	wake_up_interruptible(&info->delta_msr_wait);
	//Free the IRQ, if necessary
	if (info->xmit_buf) 
	{
	    free_page((unsigned long)info->xmit_buf);
	    info->xmit_buf = 0;
	}

	info->IER = 0;
	//disable all intrs
	byte_write(0x00, info->base + UART_IER);	

	if ( !info->tty || (info->tty->termios->c_cflag & HUPCL) )
	    info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
	byte_write(info->MCR, info->base + UART_MCR);

	//clear Rx/Tx FIFO's
	byte_write((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), info->base + UART_FCR);
	//read data port to reset things
	byte_read(info->base + UART_RX);

	if (info->tty)
	    set_bit(TTY_IO_ERROR, &info->tty->flags);

	info->flags &= ~ASYNC_INITIALIZED;
	restore_flags(flags);
}

//This routine is called to set the UART divisor registers to match
//the specified baud rate for a serial port.
static int exser_change_speed(struct exser_struct * info,
                              struct termios * old_termios)
{
	int			quot = 0;
	unsigned	cflag, cval, fcr;
	int			i;
    int         ret = 0;

//	printk("exser: change_speed\n");
	if ( !info->tty || !info->tty->termios )
	    return ret;
	cflag = info->tty->termios->c_cflag;
	if ( !(info->base) )
	    return ret;
           
#ifndef B921600
#define B921600 (B460800 + 1)
#endif 
	switch( cflag & (CBAUD | CBAUDEX) ){
        case B921600 : 
				i = 20; 
			break;
        case B460800 : 
				i = 19; 
			break;
        case B230400 : 
				i = 18; 
			break;
        case B115200 : 
				i = 17; 
			break;
        case B57600 : 
				i = 16; 
			break;
        case B38400 : 
				i = 15; 
			break;
        case B19200 : 
				i = 14; 
			break;
        case B9600 : 
				i = 13; 
			break;
        case B4800 : 
				i = 12; 
			break;
        case B2400 : 
				i = 11; 
			break;
        case B1800 : 
				i = 10; 
			break;
        case B1200 : 
				i = 9; 
			break;
        case B600 : 
				i = 8; 
			break;
        case B300 : 
				i = 7; 
			break;
        case B200 : 
				i = 6; 
			break;
        case B150 : 
				i = 5; 
			break;
        case B134 : 
				i = 4; 
			break;
        case B110 : 
				i = 3; 
			break;
        case B75 : 
				i = 2; 
			break;
        case B50 : 
				i = 1; 
			break;
        default:   
				i = 0; 
			break;
	}
	
	if (i == 15) 
	{
		if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
			//57600 bps
			i = 16; 					
		if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
			//115200 bps
			i = 17; 					
		
#ifdef ASYNC_SPD_SHI
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			i = 18;
#endif
		
#ifdef ASYNC_SPD_WARP
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			i = 19;
#endif
	}
	if (exvar_baud_table[i] == 134) 
	{
		quot = (2*info->baud_base/269);
	} 
	else if (exvar_baud_table[i]) 
	{
		quot = info->baud_base/exvar_baud_table[i];
		if (!quot && old_termios) 
		{
			//re-calculate
			info->tty->termios->c_cflag &= ~CBAUD;
			info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
			switch( info->tty->termios->c_cflag & (CBAUD | CBAUDEX) ){
			case B921600 : 
					i = 20; 
				break;
			case B460800 : 
					i = 19; 
				break;
			case B230400 : 
					i = 18; 
				break;
			case B115200 : 
					i = 17; 
				break;
			case B57600 : 
					i = 16; 
				break;
			case B38400 : 
					i = 15; 
				break;
			case B19200 : 
					i = 14; 
				break;
			case B9600 : 
					i = 13; 
				break;
			case B4800 : 
					i = 12; 
				break;
			case B2400 : 
					i = 11; 
				break;
			case B1800 : 
					i = 10; 
				break;
			case B1200 : 
					i = 9; 
				break;
			case B600 : 
					i = 8; 
				break;
			case B300 : 
					i = 7; 
				break;
			case B200 : 
					i = 6; 
				break;
			case B150 : 
					i = 5; 
				break;
			case B134 : 
					i = 4; 
				break;
			case B110 : 
					i = 3; 
				break;
			case B75 : 
					i = 2; 
				break;
			case B50 : 
					i = 1; 
				break;
			default:   
					i = 0; 
				break;
			}
			if (i == 15) 
			{
				if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
					//57600 bps
					i = 16; 					
				if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
					//115200 bps
					i = 17; 					
#ifdef ASYNC_SPD_SHI
				if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
					i = 18;
#endif
#ifdef ASYNC_SPD_WARP
				if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
					i = 19;
#endif
			}
			if (exvar_baud_table[i] == 134) 
			{
				quot = (2*info->baud_base/269);
			} 
			else if (exvar_baud_table[i]) 
			{
				quot = info->baud_base/exvar_baud_table[i];
				if(quot==0)
					quot = 1;
			} 
			else 
			{
				quot = 0;
			}
		}
		else if(quot==0)
			quot = 1;
	} 
	else 
	{
		quot = 0;
	}
	
	if (quot) 
	{
		info->MCR |= UART_MCR_DTR;
		cli();
		byte_write(info->MCR, info->base + UART_MCR);
		sti();
	} 
	else 
	{
		info->MCR &= ~UART_MCR_DTR;
		cli();
		byte_write(info->MCR, info->base + UART_MCR);
		sti();
		return ret;
	}
	//byte size and parity
	switch (cflag & CSIZE) {
	case CS5: 
			cval = 0x00; 
		break;
	case CS6: 
			cval = 0x01; 
		break;
	case CS7: 
			cval = 0x02; 
		break;
	case CS8: 
			cval = 0x03; 
		break;
	default:  
			cval = 0x00; 
		break;
	}
	if (cflag & CSTOPB)
		cval |= 0x04;
	if (cflag & PARENB)
		cval |= UART_LCR_PARITY;
	if ( !(cflag & PARODD) )
		cval |= UART_LCR_EPAR;
	if ( (info->type == PORT_8250) || (info->type == PORT_16450) ) 
	{
		fcr = 0;
	} 
	else 
	{
		fcr = UART_FCR_ENABLE_FIFO;
		switch (info->rx_trigger) 
		{
		case 1:  
			fcr |= UART_FCR_TRIGGER_1; 
			break;
		case 4:  
			fcr |= UART_FCR_TRIGGER_4; 
			break;
		case 8:  
			fcr |= UART_FCR_TRIGGER_8; 
			break;
		default: 
			fcr |= UART_FCR_TRIGGER_14;
		}
	}
	
	//CTS flow control flag and modem status interrupts
	info->IER &= ~UART_IER_MSI;
	info->MCR &= ~UART_MCR_AFE;
	if (cflag & CRTSCTS) 
	{
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
		if (info->type == PORT_16550A) 
			info->MCR |= UART_MCR_AFE;
	} 
	else 
	{
		info->flags &= ~ASYNC_CTS_FLOW;
	}
	byte_write(info->MCR, info->base + UART_MCR);
	if (cflag & CLOCAL)
		info->flags &= ~ASYNC_CHECK_CD;
	else 
	{
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}
	byte_write(info->IER, info->base + UART_IER);
	
	//Set up parity check flag
	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if (I_INPCK(info->tty))
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if ( I_BRKINT(info->tty) || I_PARMRK(info->tty) )
		info->read_status_mask |= UART_LSR_BI;
	
	info->ignore_status_mask = 0;
#if 0
	//This should be safe, but for some broken bits of hardware
	if ( I_IGNPAR(info->tty) ) {
		info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
		info->read_status_mask |= UART_LSR_PE | UART_LSR_FE;
	}
#endif
	if (I_IGNBRK(info->tty)) 
	{
		info->ignore_status_mask |= UART_LSR_BI;
		info->read_status_mask |= UART_LSR_BI;

		//If we're ignore parity and break indicators, ignore
		//overruns too.  (For real raw support).
		if (I_IGNPAR(info->tty)) 
		{
			info->ignore_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
			info->read_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
		}
	}
	cli();
	byte_write(cval | UART_LCR_DLAB, info->base + UART_LCR);	//set DLAB
	byte_write(quot & 0xff, info->base + UART_DLL);				//LS of divisor
	byte_write(quot >> 8, info->base + UART_DLM); 				//MS of divisor
	byte_write(cval, info->base + UART_LCR);					//reset DLAB
	byte_write(fcr, info->base + UART_FCR);						//set fcr
	sti();
	
	return ret;
}

//friends of exser_ioctl()
static int exser_get_serial_info(struct exser_struct * info,
								 struct serial_struct * retinfo)
{
	struct serial_struct	tmp;
	
	if ( !retinfo )
		return(-EFAULT);
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->port;
	tmp.port = info->base;
	tmp.irq = info->irq;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	tmp.hub6 = 0;
	copy_to_user(retinfo, &tmp, sizeof(*retinfo));
	return(0);
}

static int exser_set_serial_info(struct exser_struct * info,
								 struct serial_struct * new_info)
{
	struct serial_struct	new_serial;
	unsigned int		flags;
	int		retval = 0;
	
	if ( !new_info || !info->base )
		return(-EFAULT);
	copy_from_user(&new_serial, new_info, sizeof(new_serial));
	
	if ( (new_serial.irq != info->irq) ||
		(new_serial.port != info->base) ||
		(new_serial.type != info->type) ||
		(new_serial.custom_divisor != info->custom_divisor) ||
		(new_serial.baud_base != info->baud_base) )
		return(-EPERM);
	
	flags = info->flags & ASYNC_SPD_MASK;
	
	if ( !suser() ) 
	{
		if ( (new_serial.baud_base != info->baud_base) ||
					(new_serial.close_delay != info->close_delay) ||
					((new_serial.flags & ~ASYNC_USR_MASK) !=
					(info->flags & ~ASYNC_USR_MASK)) )
			return(-EPERM);
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			(new_serial.flags & ASYNC_USR_MASK));
	} 
	else 
	{

	//OK, past this point, all the error checking has been done.
	//At this point, we start making changes.....
		info->flags = ((info->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
		info->close_delay = new_serial.close_delay * HZ/100;
		info->closing_wait = new_serial.closing_wait * HZ/100;
	}
	
	if (info->flags & ASYNC_INITIALIZED) 
	{
		if (flags != (info->flags & ASYNC_SPD_MASK))
		{
			exser_change_speed(info, 0);
		}
	} 
	else
		retval = exser_startup(info);
	return(retval);
}

/* Purpose: Let user call ioctl() to get info when the UART physically
*	    is emptied.  On bus types like RS485, the transmitter must
*	    release the bus after transmitting. This must be done when
*	    the transmit shift register is empty, not be done when the
*	    transmit holding register is empty.  This functionality
*	    allows an RS485 driver to be written in user space.
*/
static int exser_get_lsr_info(struct exser_struct * info, unsigned int * value)
{
	unsigned char	status;
	unsigned int	result;
	
	cli();
	status = byte_read(info->base + UART_LSR);
	sti();
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	put_to_user(result, value);
	return(0);
}

//This routine sends a break character out the serial port.
static void exser_send_break(struct exser_struct * info, int duration)
{
	if ( !info->base )
		return;
	current->state = TASK_INTERRUPTIBLE;
	cli();
	byte_write(byte_read(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	schedule_timeout(duration);
	byte_write(byte_read(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);
	sti();
}

static int exser_get_modem_info(struct exser_struct * info,
								unsigned int * value)
{
	unsigned char	control, status;
	unsigned int	result;
	
	control = info->MCR;
	cli();
	status = byte_read(info->base + UART_MSR);
	if (status & UART_MSR_ANY_DELTA)
		exser_check_modem_status(info, status);
	sti();
	result = ((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
		((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
		((status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
		((status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
		((status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
		((status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
	put_to_user(result, value);
	return(0);
}

static int exser_set_modem_info(struct exser_struct * info, unsigned int cmd,
								unsigned int * value)
{
	int	  error;
	unsigned int arg;
	
	error = verify_area(VERIFY_READ, value, sizeof(int));
	if (error)
		return(error);
	get_from_user(arg, value);
	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->MCR &= ~UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR &= ~UART_MCR_DTR;
		break;
	case TIOCMSET:
		info->MCR = ((info->MCR & ~(UART_MCR_RTS | UART_MCR_DTR)) |
			((arg & TIOCM_RTS) ? UART_MCR_RTS : 0) |
			((arg & TIOCM_DTR) ? UART_MCR_DTR : 0));
		break;
	default:
		return(-EINVAL);
	}
	cli();
	byte_write(info->MCR, info->base + UART_MCR);
	sti();
	return(0);
}

#if 0
static int exser_get_ISA_conf(int cap, struct exser_hwconf * hwconf)
{
	int		id, i, bits;
	unsigned short	regs[16], irq;
	unsigned char	scratch, scratch2;
	
	id = exser_read_register(cap, regs);
	if (id == C158_ASIC_ID)
		hwconf->board_type = ISA_BOARD_XR788;
	else if (id == C104_ASIC_ID)
		hwconf->board_type = ISA_BOARD_XR784;
	else if (id == CI104J_ASIC_ID)
		hwconf->board_type = EXSER_BOARD_CI104J;
	else
		return(0);
	irq = regs[9] & 0x0F;
	irq = irq | (irq << 4);
	irq = irq | (irq << 8);
	if ( (irq != regs[9]) || ((id == 1) && (irq != regs[10])) ) 
	{
		return(EXSER_ERR_IRQ_CONFLIT);
	}
	if ( !irq ) 
	{
		return(EXSER_ERR_IRQ);
	}
	for (i=0; i<8; i++)
		hwconf->ioaddr[i] = (int)regs[i + 1] & 0xFFF8;
	hwconf->irq = (int)(irq & 0x0F);
	if ((regs[12] & 0x80) == 0) 
	{
		return(EXSER_ERR_VECTOR);
	}
	hwconf->vector = (int)regs[11];	
	if (id == 1)
		hwconf->vector_mask = 0x00FF;
	else
		hwconf->vector_mask = 0x000F;
	for (i=7, bits=0x0100; i>=0; i--, bits <<= 1) 
	{
		if (regs[12] & bits)
			hwconf->baud_base[i] = 921600;
		else
			hwconf->baud_base[i] = 115200;
	}
	scratch2 = byte_read(cap + UART_LCR) & (~UART_LCR_DLAB);
	byte_write(scratch2 | UART_LCR_DLAB, cap + UART_LCR);
	byte_write(0, cap + UART_EFR);
	byte_write(scratch2, cap + UART_LCR);
	byte_write(UART_FCR_ENABLE_FIFO, cap + UART_FCR);
	scratch = byte_read(cap + UART_IIR);
	if (scratch & 0xC0)
		hwconf->uart_type = PORT_16550A;
	else
		hwconf->uart_type = PORT_16450;
	if (id == 1)
		hwconf->ports = 8;
	else
		hwconf->ports = 4;
	return(hwconf->ports);
}
#endif

static void exser_setled(struct exser_hwconf * hwconf)
{
	if(hwconf->brdid == 0x1)
		byte_write(0xfe, hwconf->ioaddr[0] + LED_ADDR);
	if(hwconf->brdid == 0x2)
		byte_write(0xfc, hwconf->ioaddr[0] + LED_ADDR);
	if(hwconf->brdid == 0x3)
		byte_write(0xf8, hwconf->ioaddr[0] + LED_ADDR);
}

static void exser_clearled(struct exser_hwconf * hwconf)
{
	byte_write(0xff, hwconf->ioaddr[0] + LED_ADDR);
}

//byte_read & byte_write make code work with PCI & ISA 
static unsigned char byte_read( unsigned long addr)
{
	unsigned char cByte = 0;
#if defined(XR158_PCI) || defined(XR154_PCI)
	cByte = readb(addr);
#else
    cByte = inb(addr);
#endif
	return cByte;
}

static void byte_write( unsigned char val, unsigned long addr)
{
#if defined(XR158_PCI) || defined(XR154_PCI)
	writeb(val, addr);
#else
	outb(val, addr);
#endif
}



